import fs from 'fs';
import jsonfile from 'jsonfile';
import { Flow, Step } from 'last-hit-types';
import os from 'os';
import path from 'path';
import { EnvironmentOptions, IncludingFilter, IncludingFilters } from '../types';

type Wrapper = (step: Step) => Step;
type SimpleEnvironmentOptions = Omit<EnvironmentOptions,
	'parallel' | 'workspace' | 'includes' | 'child'>;

class Environment {
	/** environment name */
	private name: string = 'NO-ENVIRONMENT';
	/** workspace folder */
	private workspace: string;
	private urlReplaceRegexps: RegExp[] = [];
	private urlReplaceTos: string[] = [];
	private sleepAfterChange?: number;
	private slowAjaxTime?: number;
	private includes?: IncludingFilters;
	private parallel?: number;
	private child: boolean;

	private adminUrl?: string;
	private adminToken?: string;
	private adminWorkspaceId?: string | number;
	private adminTestPlanId?: string | number;

	private wrappers: Wrapper[];

	private originalOptions: EnvironmentOptions;

	constructor(options: EnvironmentOptions) {
		this.originalOptions = options;

		this.workspace = options.workspace;
		this.mergeFrom(options);

		this.includes = options.includes;

		this.parallel = this.computeParallel(options.parallel);
		this.child = options.child || false;

		this.adminUrl = options.adminUrl;
		this.adminToken = options.adminToken;
		this.adminWorkspaceId = options.adminWorkspaceId;
		this.adminTestPlanId = options.adminTestPlanId;

		this.wrappers = [ this.wrapUrl ];
	}

	getOriginalOptions(): EnvironmentOptions {
		return this.originalOptions;
	}

	mergeFrom(options: SimpleEnvironmentOptions) {
		this.name = options.name || this.name;
		if (options.urlReplaceRegexp) {
			this.urlReplaceRegexps = options.urlReplaceRegexp
				.split('&&')
				.map(text => new RegExp(text));
			this.urlReplaceTos = (options.urlReplaceTo || '').split('&&');
			this.urlReplaceTos.length = this.urlReplaceRegexps.length;
			this.urlReplaceTos = this.urlReplaceTos.map(to => (to ? to : ''));
		} else {
			this.urlReplaceRegexps = [];
			this.urlReplaceTos = [];
		}
		this.sleepAfterChange = options.sleepAfterChange || this.sleepAfterChange;
		this.slowAjaxTime = options.slowAjaxTime || this.slowAjaxTime;

		// set original options
		[
			'name',
			'urlReplaceRegexp', 'urlReplaceTo',
			'sleepAfterChange', 'slowAjaxTime',
			'parallel',
			'adminUrl'
		].forEach(
			prop => (this.originalOptions[prop] = options[prop])
		);
	}

	wrap(step: Step): Step {
		return this.getWrappers().reduce((step, wrapper) => wrapper.call(this, step), step);
	}

	private wrapUrl(step: Step): Step {
		const regexps = this.getUrlReplaceRegexps();
		if ((step as any).url && regexps) {
			(step as any).url = regexps.reduce((url, regexp, index) => {
				return url.replace(regexp, this.getUrlReplaceTos()[index] || '');
			}, (step as any).url);
		}
		return step;
	}

	private getWrappers(): Wrapper[] {
		return this.wrappers;
	}

	getName(): string {
		return this.name;
	}

	getWorkspace(): string {
		return this.workspace;
	}

	getUrlReplaceRegexps(): RegExp[] {
		return this.urlReplaceRegexps;
	}

	getUrlReplaceTos(): string[] {
		return this.urlReplaceTos;
	}

	getSleepAfterChange(): number | undefined {
		return this.sleepAfterChange;
	}

	getSlowAjaxTime(): number {
		return this.slowAjaxTime || 500;
	}

	isIncluded(storyName: string, flowName: string): boolean {
		return (
			// no filters, including all
			!this.includes ||
			this.includes.some((filter: IncludingFilter): boolean => {
				// story name must match
				// no flow name appointed or flow name matched
				return filter.story === storyName && (!filter.flow || filter.flow === flowName);
			})
		);
	}

	isExcluded(storyName: string, flowName: string): boolean {
		// TODO not supported yet, always returns false
		return false;
	}

	getParallel(): number {
		return this.parallel || 1;
	}

	isOnParallel(): boolean {
		return this.getParallel() !== 1;
	}

	getAdminUrl(): string | undefined {
		return this.adminUrl;
	}

	getAdminToken(): string | undefined {
		return this.adminToken;
	}

	getAdminWorkspaceId(): string | number | undefined {
		return this.adminWorkspaceId;
	}

	getAdminTestPlanId(): string | number | undefined {
		return this.adminTestPlanId;
	}

	private computeParallel(parallel: number = 1): number {
		parallel = Math.abs(parallel);
		if (parseInt(`${parallel}`) !== parallel) {
			return Math.round(os.cpus().length * parallel);
		} else {
			return parallel;
		}
	}

	isOnChildProcess(): boolean {
		return this.child;
	}

	exposeForSingleProcess(replacement: { includes: IncludingFilters }): EnvironmentOptions {
		const options = Object.assign({}, this.originalOptions);
		delete options.parallel;
		delete options.workspace;

		options.child = true;
		Object.keys(replacement).forEach(key => (options[key] = replacement[key]));

		return options;
	}

	expose(): SimpleEnvironmentOptions {
		const options = Object.assign({}, this.originalOptions);
		delete options.parallel;
		delete options.workspace;
		delete options.includes;
		delete options.child;
		return options;
	}

	static exposeNoop(): EnvironmentOptions {
		return {
			// name: 'NO-ENVIRONMENT'
		} as EnvironmentOptions;
	}

	isStoryExists(storyName: string): boolean {
		const dependsStoryFolder = path.join(this.getWorkspace(), storyName);
		return fs.existsSync(dependsStoryFolder) && fs.statSync(dependsStoryFolder).isDirectory();
	}

	isFlowExists(storyName: string, flowName: string): boolean {
		const dependsStoryFolder = path.join(this.getWorkspace(), storyName);
		if (!this.isStoryExists(storyName)) {
			return false;
		}
		const dependsFlowFilename = path.join(dependsStoryFolder, `${flowName}.flow.json`);
		return fs.existsSync(dependsFlowFilename) && fs.statSync(dependsFlowFilename).isFile();
	}

	readFlowFile(storyName: string, flowName: string): Flow {
		const dependsStoryFolder = path.join(this.getWorkspace(), storyName);
		const filename = path.join(dependsStoryFolder, `${flowName}.flow.json`);
		return jsonfile.readFileSync(filename);
	}
}

export default Environment;
